/*
 * Copyright Contributors to the fusa-gcc-plugin project
 *
 * SPDX-License-Identifier: LGPL-2.1-or-later
 */
#include "declhelpers.hpp"

using namespace std;

namespace declhelpers {
// Get namespace hierarchy for the given decl
//
//   std::algorithm::find_if => ["std", "algorithm"]
//
vector<string> get_namespace_hierarchy(tree decl) {
  vector<string> namespace_hierarchy = {};

  DECL_SET_IF(RECORD_TYPE, TYPE_CONTEXT);

  while (decl) {
    if (TREE_CODE(decl) == NAMESPACE_DECL)
      namespace_hierarchy.insert(namespace_hierarchy.begin(),
                                 DECL_NAME_STRING(decl));

    DECL_SET_IF_ELSE(RECORD_TYPE, TYPE_CONTEXT, DECL_CONTEXT);
  }

  return namespace_hierarchy;
}

// Get class hierarchy for the given decl
//
//   std::basic_string::allocator => ["basic_string", "allocator"]
//
vector<string> get_class_hierarchy(tree decl) {
  vector<string> class_hierarchy = {};

  decl = DECL_CONTEXT(decl);

  // Calling TYPE_CONTEXT on a translation unit is an error
  // so we exit early. This happens specificaly for C functions
  if (TREE_CODE(decl) == TRANSLATION_UNIT_DECL)
    return class_hierarchy;

  while (decl) {
    if (TREE_CODE(decl) == RECORD_TYPE) {
      tree name = TYPE_NAME(decl);

      if (name && TREE_CODE(name) == TYPE_DECL)
        class_hierarchy.insert(class_hierarchy.begin(), DECL_NAME_STRING(name));
    }

    DECL_SET_IF_ELSE(FUNCTION_DECL, DECL_CONTEXT, TYPE_CONTEXT);
  }

  return class_hierarchy;
}

// Recursively iterate the tree `t` and return all nodes
// of the FUNCTION_DECL type
vector<tree> find_function_decls(tree decl) {
  vector<tree> decls = {};

  walk_tree_fn fn = [](tree *item, int *walk_subtrees, void *context) -> tree {
    vector<tree> *_decls = static_cast<vector<tree> *>(context);

    if (!item)
      return NULL;

    if (TREE_CODE(*item) == FUNCTION_DECL)
      _decls->push_back(*item);
    else
      *walk_subtrees = 1;

    return NULL;
  };

  if (decl != NULL_TREE)
    walk_tree_without_duplicates_1(&decl, fn, (void *)&decls, NULL);

  return decls;
}
} // namespace declhelpers
